/*

FILE: prom.c - SPI boot programmer
DESCRIPTION: 

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <vs1000.h>
#include <audio.h>

void putch(register __a0 short ch);

void Halt();
__near void vs3emubreak(register __i1 bp);


typedef long s_int32;
typedef short s_int16;
typedef unsigned short u_int16;
typedef unsigned long u_int32;
typedef __fract short f_int16;
typedef __fract long f_int32;

int largeAddr = 0;

static char hex[] = "0123456789abcdef";
void puthex(unsigned int a) {
    char tmp[5];
    tmp[0] = hex[(a>>12)&15];
    tmp[1] = hex[(a>>8)&15];
    tmp[2] = hex[(a>>4)&15];
    tmp[3] = hex[(a>>0)&15];
    tmp[4] = '\0';
    puts(tmp);
}

//#define USEY(x) *((__y volatile unsigned short *)(unsigned short)(x))

#define SPI_WREN  0x06
#define SPI_WRDI  0x04
#define SPI_RDSR  0x05
#define SPI_WRSR  0x01
#define SPI_READ  0x03
#define SPI_WRITE 0x02

u_int16 SpiRead(u_int16 addr) {
    u_int16 dat;

    PERIP(SPI0_CONFIG) = SPI_CF_MASTER | SPI_CF_DLEN8 | SPI_CF_FSIDLE1;
    SpiDelay(0);
    PERIP(SPI0_CONFIG) = SPI_CF_MASTER | SPI_CF_DLEN8 | SPI_CF_FSIDLE0;
    SpiSendReceive(SPI_READ);
    PERIP(SPI0_CONFIG) = SPI_CF_MASTER | SPI_CF_DLEN16 | SPI_CF_FSIDLE0;
    SpiSendReceive(addr);
    dat = SpiSendReceive(0);
    SpiDelay(0);
    PERIP(SPI0_CONFIG) = SPI_CF_MASTER | SPI_CF_DLEN8 | SPI_CF_FSIDLE1;
    return dat;
}

u_int16 SpiStatus(void) {
    u_int16 dat;

    PERIP(SPI0_CONFIG) = SPI_CF_MASTER | SPI_CF_DLEN8 | SPI_CF_FSIDLE1;
    SpiDelay(0);
    PERIP(SPI0_CONFIG) = SPI_CF_MASTER | SPI_CF_DLEN8 | SPI_CF_FSIDLE0;
    SpiSendReceive(SPI_RDSR);
    dat = SpiSendReceive(0);
    SpiDelay(0);
    PERIP(SPI0_CONFIG) = SPI_CF_MASTER | SPI_CF_DLEN8 | SPI_CF_FSIDLE1;
    return dat;
}

u_int16 SpiWrite(u_int16 addr, u_int16 wrdat) {
    u_int16 dat;

    PERIP(SPI0_CONFIG) = SPI_CF_MASTER | SPI_CF_DLEN8 | SPI_CF_FSIDLE1;
    SpiDelay(0);
    PERIP(SPI0_CONFIG) = SPI_CF_MASTER | SPI_CF_DLEN8 | SPI_CF_FSIDLE0;
    SpiSendReceive(SPI_WREN);
    SpiDelay(0);
    PERIP(SPI0_CONFIG) = SPI_CF_MASTER | SPI_CF_DLEN8 | SPI_CF_FSIDLE1;

    SpiDelay(0);
    PERIP(SPI0_CONFIG) = SPI_CF_MASTER | SPI_CF_DLEN8 | SPI_CF_FSIDLE0;
    SpiSendReceive(SPI_WRITE);
    PERIP(SPI0_CONFIG) = SPI_CF_MASTER | SPI_CF_DLEN16 | SPI_CF_FSIDLE0;
    SpiSendReceive(addr);
    dat = SpiSendReceive(wrdat);
    SpiDelay(0);
    PERIP(SPI0_CONFIG) = SPI_CF_MASTER | SPI_CF_DLEN8 | SPI_CF_FSIDLE1;
    return dat;
}


#if 1
#define EEPROM_PAGE_SIZE_BITS 5
#define EEPROM_PAGE_SIZE (1<<EEPROM_PAGE_SIZE_BITS)

u_int16 SpiBlockWrite(u_int16 addr, register u_int16 *buf, u_int16 len) {
    PERIP(SPI0_CONFIG) = SPI_CF_MASTER | SPI_CF_DLEN8 | SPI_CF_FSIDLE1;
    SpiDelay(0);
    PERIP(SPI0_CONFIG) = SPI_CF_MASTER | SPI_CF_DLEN8 | SPI_CF_FSIDLE0;
    SpiSendReceive(SPI_WREN);
    PERIP(SPI0_CONFIG) = SPI_CF_MASTER | SPI_CF_DLEN8 | SPI_CF_FSIDLE1;
    SpiDelay(0);
    PERIP(SPI0_CONFIG) = SPI_CF_MASTER | SPI_CF_DLEN8 | SPI_CF_FSIDLE0;
    SpiSendReceive(SPI_WRITE);
    if (largeAddr) {
	SpiSendReceive(0); //24-bit address
    }
    PERIP(SPI0_CONFIG) = SPI_CF_MASTER | SPI_CF_DLEN16 | SPI_CF_FSIDLE0;
    SpiSendReceive(addr);
    while (len--) {
	SpiSendReceive(*buf);
	addr += 2;
	buf++;
    }
    SpiDelay(0);
    PERIP(SPI0_CONFIG) = SPI_CF_MASTER | SPI_CF_DLEN8 | SPI_CF_FSIDLE1;
    return addr;
}

u_int16 SpiVerify(u_int16 addr, register u_int16 *buf, u_int16 len) {
    u_int16 dat;

    PERIP(SPI0_CONFIG) = SPI_CF_MASTER | SPI_CF_DLEN8 | SPI_CF_FSIDLE1;
    SpiDelay(0);
    PERIP(SPI0_CONFIG) = SPI_CF_MASTER | SPI_CF_DLEN8 | SPI_CF_FSIDLE0;
    SpiSendReceive(SPI_READ);
    if (largeAddr) {
	SpiSendReceive(0); //24-bit address, send the top 8 bits as 0
    }
    PERIP(SPI0_CONFIG) = SPI_CF_MASTER | SPI_CF_DLEN16 | SPI_CF_FSIDLE0;
    SpiSendReceive(addr);
    SpiDelay(0);
    while (len--) {
	if (SpiSendReceive(0) != *buf++)
	    return -1;
    }
    PERIP(SPI0_CONFIG) = SPI_CF_MASTER | SPI_CF_DLEN8 | SPI_CF_FSIDLE1;
    return 0;
}

u_int16 BlockProgram(u_int16 addr, u_int16 *buf, s_int16 words) {
    register s_int16 i, len = words;
    for (i=0;i<len;) {
	register u_int16 sz = (((addr&~(EEPROM_PAGE_SIZE-1))+EEPROM_PAGE_SIZE)
			       - addr) / 2;
	if (sz > len-i)
	    sz = len-i;
	SpiBlockWrite(addr, buf+i, sz);
	i += sz;
	addr += 2*sz;
	while (SpiStatus() & 1)
	    SpiDelay(0);
    }
    return SpiVerify(addr - 2*len, buf, len);
}
#endif

void SpiInit() {
    PERIP(GPIO1_MODE) |= 15; /*XCS SCLK SI SO to perip mode */
    PERIP(SPI0_CONFIG) = SPI_CF_MASTER | SPI_CF_DLEN8 | SPI_CF_FSIDLE1;
    PERIP(SPI0_FSYNC) = 0;
    //PERIP(SPI0_CLKCONFIG) = SPI_CC_CLKDIV * (12-1);
    PERIP(SPI0_CLKCONFIG) = SPI_CC_CLKDIV * (clockX-1);
}


#define BUF_SIZE 256
void main(void) {
    FILE *fp;
    static u_int16 buf[BUF_SIZE];
    int i, len, addr = 0;

    PERIP(INT_ENABLE) = 0;
    SpiInit(); /* Set SPI clock etc */

    puthex(SpiStatus()); /* should be 0080 */
    puts(" status");

    fp = fopen("boot.img", "rb");
    if (!fp) {
	puts("Could not open boot.img!");
	((void (*)(void))(0x4000))();
    }
    while (len = fread(buf, 1, BUF_SIZE, fp)) {
#ifdef EEPROM_PAGE_SIZE
	if (BlockProgram(addr, buf, len)) {
	    putchar('x');
	    putchar('\n');
	} else {
	    putchar('.');
	    fflush(stdout);
	}
	addr += len*2;
#else
	for (i=0;i<len;i++) {
	    SpiWrite(addr, buf[i]);
#if 0
	    printf("0x%04x\n", SpiStatus());
#endif
	    if (!(addr & 15)) {
		putchar('.');
		fflush(stdout);
	    }
	    while (SpiStatus() & 1) {
		SpiDelay(0);
	    }

#if 0
	    puthex(SpiRead(addr));
#endif
#if 1
	    if (buf[i] != SpiRead(addr)) {
		puthex(addr);
		putchar(' ');
		puthex(SpiRead(addr));
		putchar('\n');
	    }
#endif
	    addr += 2;
	}
#endif
    }
    fclose(fp);

    puts("\nFinished!!\n");
    fflush(stdout);

    ((void (*)(void))(0x4000))();
}
