#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <vs1000.h>
#include <audio.h>
#include <mappertiny.h>
#include <minifat.h>
#include <codec.h>
#include <vsNand.h>
#include <player.h>
#include <usblowlib.h>

#include <dev1000.h>

#include "bookvars.h"

struct screen {
    u_int16 X, Y, invertMask;
} screen = {
    0, 0, 0
};

extern struct {
    u_int16 level;
    u_int32 checkTime;
} battery;

volatile u_int16 menuMode = 0;

//#define USE_DEBUG


#define USE_USB
#define DISPLAY
//#define PARALLEL_DISPLAY /* Must be set in makefile to affect ASM also! */
#define BATTERY_LEVEL (1<<5) /*GPIO0 pin*/
//#define AUTO_POWEROFF //disabled to allow easier USB-programming
#define POWERON_KEYWAIT 4


#ifdef DISPLAY
void ScreenLocate(register __b0 int x, register __b1 int y); /*34 words*/
void ScreenPutData(register __a1 u_int16 c); /*36 words*/

#define SPI_MASTER_8BIT_CSHI    PERIP(SPI0_CONFIG) = SPI_CF_MASTER | SPI_CF_DLEN8 | SPI_CF_FSIDLE1
#define SCREEN_SET_COMMAND_MODE PERIP(GPIO1_CLEAR_MASK) = 0x0004
#define SCREEN_SET_DATA_MODE    PERIP(GPIO1_SET_MASK) = 0x0004
#define SCREEN_SELECT   PERIP(GPIO0_CLEAR_MASK) = 0x4000 //CS2
#define SCREEN_END      PERIP(GPIO0_SET_MASK) = 0x4000 //CS2


#ifndef PARALLEL_DISPLAY
void ScreenInit(void) {
    SCREEN_SET_COMMAND_MODE;
#if 0
    SCREEN_SELECT;
    SpiSendReceive(0xe2); /*Reset*/
    SCREEN_END;
#endif
    {
	static const unsigned char oledInit[]={
	    0xae, // Display off
	    0xd3, // Display offset
	    0x00, // 'value
	    0xa8, // Multiplex ratio
	    0x3f, // 'value
	    0xad, // DC/DC on
	    0x8b, // 'value 8b
	    0x40, // Display start line
	    0xa0, // Segment remap
	    0xc8, // Scan direction
	    0x81, // Set Contrast (=brightness)
	    0x30, // 00 is minimum, ff is maximum
	    0xb0, // Page address
	    0x00, // Column address low nibble
	    0x10, // Column address high nibble
	    0xa4, // Entire display on
	    0xa6, // Normal (not inverted) display
	    0xaf, // Display on
	    0xd8, // Power Save
	    0x05  // Low power mode
	};
	register const unsigned char *c = oledInit;
	register int i;
	SCREEN_SELECT;
	for (i=0; i<sizeof(oledInit); i++) {
	    SpiSendReceive(*c++);
	}
	SCREEN_END;
    }
}
#endif

#if 1
void ScreenLocate(register __b0 int x, register __b1 int y);
#else
// Set x,y pointer of display controller
void ScreenLocate(register __b0 int x, register __b1 int y) { /*34 words*/
  SCREEN_SELECT; //CS low
#ifdef PARALLEL_DISPLAY /* Must be set in makefile to affect ASM also! */
  NandPutCommand((screen.Y = y) | 0xb0);
  NandPutCommand((screen.X = x) & 0x0f);
  NandPutCommand((x >> 4) | 0x0010);
#else
  SCREEN_SET_COMMAND_MODE;
  SpiSendReceive((screen.Y = y) | 0xb0);
  SpiSendReceive((screen.X = x) & 0x0f);
  SpiSendReceive((x >> 4) | 0x0010);
#endif
  SCREEN_END;
}
#endif

#if 1
void ScreenPutData(register __a1 u_int16 c);
#else
// Write image data octet to display controller
void ScreenPutData(register __a1 u_int16 c) { /*36 words*/
  SCREEN_SELECT; //CS low
#ifdef PARALLEL_DISPLAY /* Must be set in makefile to affect ASM also! */
  NandPutAddressOctet(c ^ screen.invertMask);
#else
  SCREEN_SET_DATA_MODE;
  SpiSendReceive(c ^ screen.invertMask);
#endif
  SCREEN_END; //CS high
  if (++screen.X > 131) {
      ScreenLocate(0, ++screen.Y);
  }
}
#endif
#if 1
void ScreenClear();
#else
void ScreenClear() { /*29 words*/
    register u_int16 i;
    ScreenLocate(0, 0);
    for (i=0;i<8*132;i++)
	ScreenPutData(0); /*ScreenPutData() handles the line wraps*/
    ScreenLocate(0, 0);
}
#endif
#endif/*DISPLAY*/



extern struct FsNandPhys fsNandPhys;
extern struct FsPhysical *ph;
extern struct FsMapper *map;
extern struct Codec *cod;
extern struct CodecServices cs;
extern u_int16 codecVorbis[];

#ifdef BATTERY_LEVEL

/* Now self-contained, can be loaded on-demand.. */
u_int16 CheckBatteryLevel(void) { /*65 words*/
    register u_int16 i;
    register u_int16 mode = PERIP(GPIO0_MODE);

    PERIP(GPIO0_MODE) = mode & ~BATTERY_LEVEL;
    PowerSetVoltages(&voltages[voltCoreUser]);

    /* Wait a bit... about 200us */
    for (i=0;i<30000;i++)
	PERIP(GPIO0_IDATA);
    if ((PERIP(GPIO0_IDATA) & BATTERY_LEVEL)) {
	/*VBATT > VREF(=0.36*AVDD)
	  Raise VREF and try again. */
	i = voltages[voltAnaUser];
	if (voltages[voltAnaUser] < 31)
	    voltages[voltAnaUser]++;
    } else {
	/*VBATT < VREF(=0.36*AVDD) */
	if (voltages[voltAnaUser] > 12)
	    voltages[voltAnaUser]--;
	i = voltages[voltAnaUser];
    }
    PERIP(GPIO0_MODE) = mode;
    PowerSetVoltages(&voltages[voltCorePlayer]);
    return i; 
}

/*
  Full measurement of batter level. Takes time, so it should not
  be called from IdleHook().
 */
#if 0
/* Used in startup program */
void IMeasureBatteryLevel(void) { /*29 words*/
}
#endif
#endif/*BATTERY_LEVEL*/



#if 1
void InitKeyEventHandler(enum keyEvent event) { /*140 words*/
    if (event != ke_ff_off)
	menuMode = 1;
}
#endif



#ifdef SAVE_SETTINGS
u_int16 CalcCheck(register __y __i2 u_int16 *p) {
    register int i;
    register u_int16 chk = 0xffffU;
    for (i=0;i<sizeof(settings);i++)
	chk ^= *p++;
    return chk;
}
#endif/*SAVE_SETTINGS*/



int MenuInit(void) { /*51 words*/
    static const u_int32 mnuFiles[] = {	FAT_MKID('M','N','U'), 0 };
    minifatInfo.supportedSuffixes = &mnuFiles[0];
    /* MENU.MNU */
    if (OpenFileBaseName("\pMENU    ") != 0xffffU) {
	menuStart = minifatFragments[0].start & 0x7fffffffUL;
	minifatInfo.supportedSuffixes = supportedFiles;
	return 0;
    }
    minifatInfo.supportedSuffixes = supportedFiles;
    return -1;
}


#if 1
void InitUserInterfaceIdleHook(void) { /*94 words*/
    if (uiTrigger) {
	uiTrigger = 0;
	NandPutCommand(0); /*Bring data bus low*/
	KeyScan7();
    }
}
#endif

#ifdef USE_USB
void MyInitUSBDescriptors(u_int16 initDescriptors) {
    RealInitUSBDescriptors(1 /* always mass storage */);
}
#endif


const struct KeyMapping keyMaps[3][16] = { /*do not change dimensions*/
    {
	{KEY_POWER, ke_pauseToggle},
	{KEY_LONG_PRESS|KEY_POWER, (int)ke_stop},
	{KEY_1, ke_volumeDown},
	{KEY_LONG_PRESS|KEY_1, ke_volumeDown},
	{KEY_2, ke_volumeUp},
	{KEY_LONG_PRESS|KEY_2, ke_volumeUp},
	{KEY_3, ke_previous},
	{KEY_LONG_PRESS|KEY_3, ke_rewind},
	{KEY_4, ke_next},
	{KEY_LONG_PRESS|KEY_4, ke_ff_faster},
	{KEY_LONG_PRESS|KEY_RELEASED, ke_ff_off},
	{KEY_5, ke_earSpeaker},
	{KEY_LONG_PRESS|KEY_5, (u_int16)(ke_menu|KEY_LONG_ONESHOT)},
#if 1
	{KEY_6, (int)ke_bookmark},
	{KEY_LONG_PRESS|KEY_6, (u_int16)(ke_hold|KEY_LONG_ONESHOT)},
#endif
	{0, ke_null}
    }, {
	{KEY_POWER, (int)ke_play},
	{KEY_LONG_PRESS|KEY_POWER, (int)ke_powerOff},
	{KEY_1, (int)ke_down},
	{KEY_LONG_PRESS|KEY_1, (int)ke_down},
	{KEY_2, (int)ke_up},
	{KEY_LONG_PRESS|KEY_2, (int)ke_up},
	{KEY_3, (int)ke_parent},
	{KEY_LONG_PRESS|KEY_3, (u_int16)(ke_parent|KEY_LONG_ONESHOT)},
	{KEY_4, (int)ke_select},
	{KEY_LONG_PRESS|KEY_4, (u_int16)(ke_select|KEY_LONG_ONESHOT)},
	{KEY_5, (int)ke_select/*(int)ke_earSpeaker*/},
	{KEY_LONG_PRESS|KEY_5, (u_int16)(ke_return|KEY_LONG_ONESHOT)},
#if 1
	{KEY_6, (int)ke_bookmark},
	{KEY_LONG_PRESS|KEY_6, (u_int16)(ke_hold|KEY_LONG_ONESHOT)},
#endif
	{0, ke_null}
    }, { /* Keys locked mode */
	//{KEY_LONG_PRESS|KEY_POWER, (int)ke_powerOff},
#if 1
	{KEY_LONG_PRESS|KEY_6, (u_int16)(ke_hold|KEY_LONG_ONESHOT)},
#else
	{KEY_LONG_PRESS|KEY_POWER, (u_int16)(ke_hold|KEY_LONG_ONESHOT)},
#endif
	{0, ke_null}
    }
};

void main(void) { /*264 words*/
    register u_int32 time;

    //battery.level = 12;
    battery.checkTime = 0;
    clockX = 6; /*3.0x*/
    /* set cs.sampleRate to non-zero because we use cs.Output() */
    cs.sampleRate = 44100U;

    /* Turn off after 5 minutes (300secs) in pause mode. */
    player.offDelay = 300/5; 
    //player.offDelay = 20/5; 

    {
	/* The display needs the IO voltage to be about 2.6V */
	static const __y myVoltages[] = {
	    /* Core I/O Analog */
	    22, 27, 12, /* 2.2 3.3 2.8 player */
	    27, 27, 30, /* 2.5 3.3 3.6 USB */
	    14, 17,  0, /* 1.8 2.6 3.6 (USB) suspend  18=2.65*/
	    22, 27, 12  /* 2.2 3.3 2.8 user -- used for battery measurement*/
	};
	memcpyYX(voltages, myVoltages, sizeof(myVoltages));
	/* battery measurement does this.. */
	//PowerSetVoltages(&voltages[voltCorePlayer]);
    }

    /* Set the leds after nand-boot! */
    PERIP(GPIO1_ODATA) |=  LED1;   /* POWER led on */
    PERIP(GPIO1_DDR)   |=  LED1|LED2;   /* SI and SO to outputs */
    PERIP(GPIO1_MODE)  &= ~(LED1|LED2); /* SI and SO to GPIO */

#ifdef USE_USB
    SetHookFunction((u_int16)InitUSBDescriptors, MyInitUSBDescriptors);
#endif

    player.volumeOffset = -24;
    player.pauseOn = 0;
    //player.randomOn = 0;
    keyOld = KEY_POWER;
    keyOldTime = -32767; /* ignores the first release of KEY_POWER */

    map = FsMapTnCreate(ph, 0); /* Create tiny Mapper */

#ifdef POWERON_KEYWAIT /*70 words! -- only have it in the startup program */
    /* Wait until powerbutton up */
 more:
    while ((USEX(SCI_STATUS) & SCISTF_REGU_POWERBUT)) {
	if (timeCount > (POWERON_KEYWAIT*TIMER_TICKS/16)) {
	    PERIP(GPIO1_ODATA) |= LED1;
	    break;
	}
	if ((timeCount & 0xc)==0) {
	    PERIP(GPIO1_ODATA) |= LED1;
	} else {
	    PERIP(GPIO1_ODATA) &= ~LED1;
	}
    }
    /* If < 1.5sec, power off */
    if (timeCount < (POWERON_KEYWAIT*TIMER_TICKS/16)) {
	PERIP(GPIO1_ODATA) &= ~LED1;
	/* wait at least 1 sec before poweroff
	   to make certain power is properly ON. */
	while (timeCount < TIMER_TICKS) {
	    if ((USEX(SCI_STATUS) & SCISTF_REGU_POWERBUT))
		goto more;
	    Sleep();
	}
	PowerOff(); /* no settings changed */
    }
    PERIP(GPIO1_ODATA) |= LED1;
#endif

    settings.numBookMarks = 0;
    settings.backLightTime = 20*TIMER_TICKS; /* 20 seconds */
    settings.volume = 17;
#ifdef SAVE_SETTINGS
    {
	int n = ph->eraseBlockSize;
	__y u_int32 block;
	if (n < 256) {
	    block = 256-n; /* use the last erase block before mapper */
	} else {
	    ph->eraseBlocks--; /* Reserve one erase block for our settings */
	    block = (u_int32)ph->eraseBlocks * n; /* use the last erase block*/
	}
	while (n--) {
	    register int i = 0;
	    ph->Read(ph, block, 1, minifatBuffer, NULL);
	    while (i < 256) {
		if (minifatBuffer[i] == 0xffffU)
		    break; /* first unused entry */
		memcpyXY(&settings, minifatBuffer+i, sizeof(settings));
		if (CalcCheck((__y void *)&settings) == 0) {
#if 0
		    {
			register int j;
			for (j=0;j<sizeof(settings);j++)
			    puthex(((__y u_int16 *)&settings)[j]);
			puthex(block);
			puthex(i);
			puts("=got");
		    }
#endif
		    goto gotone;//break;
		}
		i += sizeof(settings);
	    }
	    block++;
	}
    gotone:
	;
    }
#endif
#ifdef USE_DEBUG
    puthex(settings.numBookMarks);
    puts("=numBookMarks");
#endif

    player.volume = settings.volume;
    if (settings.backLightTime) {
#ifdef PARALLEL_DISPLAY
	PERIP(GPIO1_ODATA) |= LED2;
#else
	PERIP(GPIO1_DDR) |= (1<<4); /*TX*/
	PERIP(GPIO1_CLEAR_MASK) = (1<<4);
	PERIP(GPIO1_MODE) &= ~(1<<4); /*TX low*/
#endif
    }
    PlayerVolume();

#ifdef DISPLAY

#ifdef PARALLEL_DISPLAY /* Must be set in makefile to affect ASM also! */
    /* Wait until XCS high (display /RES) */
#if 0
    do {
	Sleep();
    } while ((PERIP(GPIO1_IDATA) & 1) == 0);
#else
    PERIP(GPIO1_ODATA) &= ~1;   /* XCS down */
    PERIP(GPIO1_DDR)   |=  1;   /* XCS to output */
    PERIP(GPIO1_MODE)  &= ~1;   /* XCS to GPIO */
    BusyWait10();
    PERIP(GPIO1_ODATA) |=  1;   /* XCS on */
#endif
    PERIP(GPIO0_ODATA) |= 0x4000; //CS2 high
    PERIP(GPIO0_DDR)   |= 0x4000; //CS2 is Output
    PERIP(GPIO0_MODE)  &= ~0x4000; //CS2 is GPIO (Display Chip Select)
    do {
	Sleep();
    } while ((PERIP(GPIO1_IDATA) & 1) == 0);
#else
    PERIP(SPI0_FSYNC) = -1; // keep SPI CS high during transfer!
    SPI_MASTER_8BIT_CSHI;
    PERIP(SPI0_CLKCONFIG) = SPI_CC_CLKDIV * (2-1); // Spi clock mclk/4 = max 12 MHz
    PERIP(GPIO1_MODE) |= 0x1f; /* enable SPI pins */

    PERIP(GPIO0_MODE) &= ~0x4000; //CS2 is GPIO (Display Chip Select)
    PERIP(GPIO0_DDR)  |= 0x4000; //CS2 is Output

    PERIP(GPIO1_MODE) &= ~0x0004; //SI is GPIO (Display Command/Data)
    PERIP(GPIO1_DDR)  |= 0x0004; //SI is Output
#endif


#ifdef USE_DEBUG
    PERIP(NFLSH_CTRL) |= 31; /* max waits */
    if (0)
    {
	int i;
	PERIP(GPIO0_MODE) = 0;
	PERIP(GPIO0_DDR)  = ~(0x4000|(1<<10)|(1<<8));//0xffff;
	PERIP(GPIO1_MODE) = 0x30; /*keep TX/RX*/
	for (i=0;i<20;i++) {
	    if (i < 16)
		PERIP(GPIO0_ODATA) = (1<<i);
	    else
		PERIP(GPIO1_ODATA) = (1<<(i-16));

	    puthex(PERIP(GPIO0_MODE));
	    puthex(PERIP(GPIO0_DDR));
	    puthex(PERIP(GPIO0_ODATA));
	    puthex(PERIP(GPIO0_IDATA));
	    puts("=GPIO0 MODE,DDR,ODATA,IDATA");
	    puthex(PERIP(GPIO1_MODE));
	    puthex(PERIP(GPIO1_DDR));
	    puthex(PERIP(GPIO1_ODATA));
	    puthex(PERIP(GPIO1_IDATA));
	    puts("=GPIO1 MODE,DDR,ODATA,IDATA");
	}
    }
#endif

#ifdef PARALLEL_DISPLAY /* Must be set in makefile to affect ASM also! */
    SCREEN_SELECT; //CS low
    {
	static const u_int16 init[] = {
	    0xaf, // Display ON
	    0x2f, // Booster, Regulator, Follower ON
	    0xa4, // Normal display (0xa5=All pixels on)
	    0xc8, /* COM scan direction: reverse */
	    0xa1, /* SEG output correspondance: reverse */
	    0
	};
	register const u_int16 *p = init;
	while (*p) {
	    NandPutCommand(*p++);
	}
    }
    SCREEN_END;
#else
    ScreenInit();
#endif
#endif

 splash:
#ifdef USE_DEBUG
    puts("checking usb");
#endif
#ifdef USE_USB
    if (USBIsAttached())
	goto usb;
#endif

#ifdef USE_DEBUG
    puts("checking battery");
#endif
#ifdef BATTERY_LEVEL
    {
	register int i;
	for (i=0;i<32;i++) {
	    battery.level = CheckBatteryLevel();
	}
#ifdef AUTO_POWEROFF
	if (battery.level <= 14) {
	    PowerOff();
	}
#endif
    }
#endif

    memcpy(USB.pkt.payload+132, &keyMaps[0][0], sizeof(keyMaps));

    /* display the splash screen */
    /* Try to init FAT. */
    if (InitFileSystem() == 0) {
	static const u_int32 scrFiles[] = {
	    FAT_MKID('S','C','R'), 0
	};
#ifdef USE_DEBUG
	puts("InitFileSystem ok");
#endif
	minifatInfo.supportedSuffixes = &scrFiles[0];
	/* Install default main menu graphics:
	   MAIN MENU
	   ==========================
	*/
	memset(USB.pkt.payload, 0, 132/2);
	memcpy(USB.pkt.payload,
	       "\xfffe\x0c18\x0cfe\xff00" "\xfeff\x1b1b\xfffe"
	       "\x0081\xffff\x8100" "\xfffe\x1c38\x7fff"
	       "\x0000\x0000"
	       "\xfffe\x0c18\x0cfe\xff00" "\xffff\xdbdb\xc3c3"
	       "\x00ff\xfe1c\x387f\xff00" "\x7fff\xc0c0\xff7f"
	       "\x0000", 30);
	memset(USB.pkt.payload+132/2, 0x0a0a, 132/2);
	/* If TOPLINE.SCR exists, replace the default graphics */
	if (OpenFileBaseName("\pMAINMENU") != 0xffffU) {
#ifdef USE_DEBUG
	    puts("MAINMENU found");
#endif
	    ReadFile(USB.pkt.payload, 0, 2*132);
	}
	/* if SCREEN.SCR exists, show it and wait before going to main prog */
	if (OpenFileBaseName("\pSCREEN  ") != 0xffffU) {
	    register int i;
	    register u_int16 *p = mallocAreaX;
#ifdef USE_DEBUG
	    puts("SCREEN found");
#endif
	    ReadFile(mallocAreaX, 0, 8*132);
	    ScreenLocate(0, 0);

screen.invertMask = 0xff;

	    {
		register u_int16 *d = mallocAreaX;
		for (i=0;i<8*132/2;i++) {
		    ScreenPutData((s_int16)*d>>8);
		    ScreenPutData(*d++);
		}
	    }
screen.invertMask = 0;

	    time = timeCount + 10 * TIMER_TICKS;
	    SetHookFunction((u_int16)KeyEventHandler, InitKeyEventHandler);
#if 1
	    SetHookFunction((u_int16)IdleHook, InitUserInterfaceIdleHook);
#endif

#ifdef USE_DEBUG
	    puts("Waiting for action..");
#endif
	    /* can not use low-power pause mode, because of the timeout */
	    do {
#ifdef USE_DEBUG
		//puts("LoadCheck");
#endif
		LoadCheck(&cs, 32); /* decrease or increase clock */
		memset(tmpBuf, 0, sizeof(tmpBuf));
#ifdef USE_DEBUG
		//puts("AudioOutputSamples");
#endif
		AudioOutputSamples(tmpBuf, sizeof(tmpBuf)/2);
		if (timeCount > time)
		    menuMode = 1;
	    } while (menuMode != 1);
	    /* Restore hooks because they will be in different addresses
	       in the main program. */
	    SetHookFunction((u_int16)KeyEventHandler, RealKeyEventHandler);
	    SetHookFunction((u_int16)IdleHook, UserInterfaceIdleHook);
#ifdef USE_DEBUG
	    puts("leaving");
#endif
	}

	minifatInfo.supportedSuffixes = supportedFiles;

    } else {
#ifdef USE_DEBUG
	puts("InitFileSystem failed");
#endif
	ScreenClear();
	/* if not valid FAT, wait for USB */
	while (1) {
	usb:
#ifdef USE_USB
	    /* If USB is attached, call our version of the MassStorage()
	       handling function. */
#ifdef PARALLEL_DISPLAY
	    PERIP(GPIO1_DDR) &= ~LED2;
#else
	    PERIP(GPIO1_MODE) |= (1<<4); /*TX high*/
#endif
	    if (USBIsAttached()) {
		PERIP(GPIO1_ODATA) &= ~LED1; /* Indicate USB */
		MassStorage();
		goto splash; /* init */
	    }
#ifdef PARALLEL_DISPLAY
	    PERIP(GPIO1_DDR) |= LED2;
#else
	    PERIP(GPIO1_MODE) &= ~(1<<4); /*TX low*/
#endif
#else
	    ;
#endif
	}
    }

    player.pauseOn = -1; /* stop mode */
    if (MenuInit()) {
	/* no menu found */
	goto usb;
    }
    voltages[voltAnaUser] = 15; /* setup battery monitor */

    /* Load the main program. */
    while (1) {
	/*TODO: save and read image with ECC enabled */
	((struct FsNandPhys *)ph)->waitns += 40;
	ph->Read(ph, 6144/512, 8192/512, mallocAreaX, NULL/*minifatBuffer*/);
	((struct FsNandPhys *)ph)->waitns -= 40;
	BootFromX(mallocAreaX);
	((struct FsNandPhys *)ph)->waitns += 20;
    }
}
